#!/usr/bin/perl -wT
#
# Copyright (c) 2004  Peter Pentchev
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
# OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
# HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
# OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
# SUCH DAMAGE.

# $Ringlet: perl/mail/qmail-msglog/qmail-msglog.pl,v 1.10 2004/10/19 16:37:28 roam Exp $

use strict;

use Config::IniFiles;
use DBI;
use Fcntl;
use IO::File;
use IO::Pipe;
use POSIX qw/:sys_wait_h/;
use Unix::Syslog qw/:macros :subs/;

my $debug = 1;
my ($sender, @recip) = (undef);
my $QERR = 111;

sub qm_debug($);
sub qm_die($ $);
sub qm_notify($);
sub qm_warn($);
sub qm_syslog($ $);

sub read_config($);

sub read_envelope();
sub decode_envelope($);

sub process_hdr($ $ $);

sub parse_wait($);

sub db_start($);
sub db_finish($ $);
sub db_store($ $);

my $cfgpath = '/var/qmail/control/msglog';
my $cfgsection = 'msglog';

my %cfg = (
	# Database connection settings
	'db_driver'	=> 'mysql',
	'db_conn'	=> 'database=maillog',
	'db_schema'	=> 'maillog.',
	'db_user'	=> 'maillog',
	'db_pass'	=> 'furrblah',
	# Here's a PostgreSQL sample
	# 'db_driver'	=> 'Pg',
	# 'db_conn'	=> 'dbname=maillog',
	# 'db_schema'	=> 'maillog.',
	# 'db_user'	=> 'maillog',
	# 'db_pass'	=> '',

	# Message data to store
	'msg_hdrs'	=> 'message-id id subject subject',
	'msg_env'	=> 'TCPLOCALHOST TCPLOCALIP TCPREMOTEHOST TCPREMOTEIP',

	# Chained qmail-queue settings
	'qmailq_next'	=> '/var/qmail/bin/qmail-queue-after-msglog /var/qmail/bin/qmail-queue',
	'qmailq_path'	=> '/var/qmail/bin:/bin:/sbin:/usr/bin:/usr/sbin:/usr/local/bin:/usr/local/sbin',
);

my %store_headers = (
	'message-id'	=> 'id',
	'subject'	=> 'subject',
);

MAIN:
{
	my ($env, $xenv);
	my ($dbh);
	my ($exitcode);
	my %stats = (
		'id'		=> undef,
		'subject'	=> undef,
		'length'	=> 0,
		'hdrlen'	=> 0,
	);
	my ($inhdr, $chdr, $cdata) = (1, undef, '');
	my ($pid, $pipe1, $pipe2);
	
	$exitcode = 0;

	# Safely skipped: open the syslog channel
	eval {
		openlog('qmail-msglog', LOG_PID, LOG_MAIL);
	};
	qm_warn("Opening the syslog socket: $@") if ($@);

	# Safely skipped: read the configuration data
	eval {
		read_config($cfgpath);
	};
	qm_warn("Reading the configuration file $cfgpath: $@") if ($@);

	qm_notify("Processing a new message");
	
	# Open the database connection
	eval {
		my $data;
		$data = join(':', 'dbi', $cfg{'db_driver'}, $cfg{'db_conn'});
		qm_debug("About to connect with '$data'") if $debug;
		$dbh = DBI->connect($data, $cfg{'db_user'}, $cfg{'db_pass'},
		    { RaiseError => 1 });
		qm_debug("dbh is now $dbh\n") if $debug;
	};
	qm_warn("Setting up the database connection: $@") if ($@);

	# Store anything we need at the start
	eval {
		db_start($dbh) if defined($dbh);
		qm_debug("Got through db_start") if $debug;
	};
	qm_warn("Storing the initial data: $@") if ($@);

	# Pass the message through
	eval {
		my ($line, $l);

		qm_debug("About to create the pipes") if $debug;
		$pipe1 = new IO::Pipe;
		qm_debug("Got the first one: $pipe1") if $debug;
		$pipe2 = new IO::Pipe;
		qm_debug("Created the pipes - second one $pipe2, about to fork") if $debug;

		if ($pid = fork()) {
			qm_debug("parent, pid is $pid") if $debug;
			$pipe1->writer();
			$pipe1->autoflush(1);
			$pipe2->writer();
			$pipe2->autoflush(1);

			# "Save as many as you can..."
			qm_debug("parent waiting to read") if $debug;
			while (defined($line = <>)) {
				$l = $line;
				$l =~ s/[\r\n]*$//;
				if ($inhdr) {
					qm_debug("parent read $l") if $debug;
					if ($l eq '') {
						process_hdr($chdr, $cdata, \%stats);
						$chdr = undef;
						qm_debug("Whole header") if $debug;
						$inhdr = 0;
					} elsif ($l =~ /^\s+(.*)/) {
						$cdata .= ' '.$1;
					} else {
						process_hdr($chdr, $cdata, \%stats);
						($chdr, $cdata) = split(/:\s+/, $l, 2);
					}
				}
				process_hdr($chdr, $cdata, \%stats) if $inhdr;

				print $pipe2 $line or last;

				$stats{'hdrlen'} += length($line) if $inhdr;
				$stats{'length'} += length($line);
			}
			close($pipe2);
		} elsif (defined($pid)) {
			qm_debug("child and stuff") if $debug;
			$pipe1->reader();
			$pipe2->reader();

			close(STDOUT);
			fcntl($pipe1, F_DUPFD, 1);
			close($pipe1);

			close(STDIN);
			fcntl($pipe2, F_DUPFD, 0);
			close($pipe2);

			$ENV{'PATH'} = $cfg{'qmailq_path'};
			delete @ENV{'IFS', 'CDPATH', 'ENV', 'BASH_ENV'};
			foreach my $q (split(/\s+/, $cfg{'qmailq_next'})) {
				qm_debug("Trying to execute $q") if $debug;
				local $SIG{'__WARN__'} = sub { };
				exec { $q } ($q) or qm_warn(
				    "Failed to execute $q: $!");
			}
			qm_die($QERR, "Failed to execute any of $cfg{qmailq_next}");
		} else {
			qm_die($QERR, "Creating the child queue process: $!");
		}
	};
	qm_warn("Passing the message through: $@") if ($@);

	# Critical: read the message envelope
	eval {
		$env = read_envelope();
		if ($debug) {
			$xenv = $env;
			$xenv =~ y/\0/#/;
			chomp $xenv;
			qm_debug("Read envelope: $xenv");
		}
		($sender, @recip) = decode_envelope($env);
		qm_notify("New message from $sender to ".join(', ', @recip));
	};
	qm_die($QERR, "Reading the envelope: $@") if ($@);

	# Send the envelope info to the child
	eval {
		print $pipe1 "$env\n" or qm_die($QERR,
		    "Could not send the envelope info to the child: $!");
		close $pipe1;

		if (waitpid($pid, 0) == -1) {
			qm_die($QERR, "Could not wait for the child to finish: $!");
		} else {
			my $waitcode = $?;
			my ($exitcode, $msg) = parse_wait($waitcode);

			qm_die($exitcode, $msg) if ($exitcode != 0);
		}
	};
	qm_die($QERR, "Sending the envelope info to the qq child: $@") if ($@);

	# Complete the database work
	if (defined($dbh)) {
		eval {
			qm_debug("About to db_finish") if $debug;
			db_finish($dbh, \%stats);
			qm_debug("db_finish returned") if $debug;
		};
		qm_warn("Storing the message data: $@") if ($@);

		# Silence is golden...
		eval {
			qm_debug("About to disconnect") if $debug;
			$dbh->disconnect();
			qm_debug("Disconnected successfully") if $debug;
		};
	}
}

sub read_config($)
{
	my ($fname) = @_;
	my ($io, $c, $p);

	$io = new IO::File;
	if (!$io->open("< $fname")) {
		qm_warn("Opening the config file $fname: $!\n");
		return;
	}
	eval {
		$c = Config::IniFiles->new(-file => $io);
	};
	qm_warn("Reading the config file $fname: $@") if $@;
	$io->close();
	return unless $c;
	foreach $p ($c->Parameters($cfgsection)) {
		$cfg{$p} = $c->val($cfgsection, $p);
	}
}

sub read_envelope()
{
	my ($env_fh, $recipline);
	
	$env_fh = IO::File->new();
	if (!$env_fh->fdopen(1, 'r')) {
		qm_die($QERR, "Could not open the message envelope");
	}
	$recipline = <$env_fh>;
	$env_fh->close();
	if (!defined($recipline)) {
		qm_die($QERR, "Could not read the message envelope")
	}
	return $recipline;
}

sub decode_envelope($)
{
	my ($env) = @_;
	my ($sender, @recip);

	if ($env !~ /^F([^\0]*)\0T(.*?)\0\0/) {
		qm_die($QERR, "Invalid format for the envelope address");
	}
	($sender, @recip) = ($1, split("\0T", $2));
	qm_debug("Sender: $sender\n") if $debug;
	if (@recip < 1) {
		qm_die($QERR, "Not a single recipient found...");
	}
	qm_debug('Recipients: '.join(', ', @recip)."\n") if $debug;
	return ($sender, @recip);
}

sub process_hdr($ $ $)
{
	my ($hdr, $data, $stat) = @_;
	my $field;

	return unless defined($hdr);
	$field = $store_headers{lc $hdr};
	$stat->{$field} = $data if defined($field);
}

sub parse_wait($)
{
	my ($waitcode) = @_;
	my ($exitcode, $msg) = ($QERR, '');
	
	if (WIFEXITED($waitcode)) {
		$exitcode = WEXITSTATUS($waitcode);
		if ($exitcode != 0) {
			$msg = "Queue child died unhappy - $exitcode";
		}
	} elsif (WIFSTOPPED($waitcode)) {
		$msg = "Queue child stopped by signal ".WSTOPSIG($waitcode).", bad bad bad";
	} elsif (WIFSIGNALED($waitcode)) {
		$msg = "Queue child killed by signal ".WTERMSIG($waitcode);
	} else {
		$msg = "Queue child died unhappy, waitpid() returned $waitcode";
	}
	return ($exitcode, $msg);
}

sub db_start($)
{
	# In a future version, here we query the database to determine
	# whether this user needs a full maillog.
}

sub db_finish($ $)
{
	my ($dbh, $stat) = @_;

	db_store($dbh, $stat);
}

sub db_store($ $)
{
	my ($dbh, $stat) = @_;
	my ($p, $q, $r, $res, $err, $i, $id, $val);

	$q = 'INSERT INTO '.$cfg{'db_schema'}.'mail_msg(tstamp, sender, msg_id, subject, length, hdrlen) VALUES(NOW(), ?, ?, ?, ?, ?)';
	qm_debug("about to $q");
	if (!($p = $dbh->prepare($q))) {
		die("Preparing the message summary query: ".
		    $dbh->errstr()."\n");
	}
	qm_debug("About to execute mail_msg") if $debug;
	qm_debug(sprintf('Values: sender %s, id %s, subj %s, len %s, hdrlen %s', $sender, $stat->{'id'} || '', $stat->{'subject'} || '', $stat->{'length'}, $stat->{'hdrlen'}));
	if (!$p->execute($sender,
	    $stat->{'id'} || '', $stat->{'subject'} || '',
	    $stat->{'length'}, $stat->{'hdrlen'})) {
		$err = $dbh->errstr();
		$p->finish();
		die("Storing the message summary: $err\n");
	}
	$p->finish();

	$id = db_fetch_insert_id($dbh);
	if (!defined($id) || $id !~ /^(\d+)$/) {
		die("Invalid SQL insert id $id\n");
	}
	qm_debug("Got SQL insert id '$id'\n");

	$q = "INSERT INTO $cfg{db_schema}mail_recip(msg_id, idx, recip) VALUES($id, ?, ?)";
	qm_debug("about to $q");
	if (!($p = $dbh->prepare($q))) {
		die("Preparing the message recipients query: ".
		    $dbh->errstr()."\n");
	}
	$i = 1;
	foreach $r (@recip) {
		qm_debug("- about to execute recip $i $r") if $debug;
		if (!$p->execute($i++, $r)) {
			qm_warn("Inserting msg_id $id, idx $i, recipient $r: ".$dbh->errstr());
		}
	}
	$p->finish();

	$q = "INSERT INTO $cfg{db_schema}mail_env(msg_id, idx, name, value) VALUES($id, ?, ?, ?)";
	qm_debug("about to $q");
	if (!($p = $dbh->prepare($q))) {
		die("Preparing the message environment query: ".
		    $dbh->errstr()."\n");
	}
	$i = 1;
	foreach $r (split(/\s+/, $cfg{'msg_env'})) {
		if (defined($ENV{$r})) {
			# Yikes.  Untaint this, it ain't going anywhere.
			$ENV{$r} =~ /^(.*)$/;
			$val = $1;
		} else {
			$val = '';
		}
		qm_debug("- about to execute env $r $val") if $debug;
		if (!$p->execute($i++, $r, $val)) {
			qm_warn("Inserting msg_id $id, idx $i, name $r, value $val: ".$dbh->errstr());
		}
	}
	$p->finish();
	qm_debug("db_store exiting") if $debug;
}

sub db_fetch_insert_id($)
{
	my ($dbh) = @_;

	if ($cfg{'db_driver'} eq 'mysql') {
		return $dbh->{'mysql_insertid'};
	} elsif ($cfg{'db_driver'} eq 'Pg') {
		my ($q, $res, $p, $r);

		$q = "SELECT currval('$cfg{db_schema}mail_msg_id'::text)::integer";
		qm_debug("about to $q");
		if (!($p = $dbh->prepare($q))) {
			die("Preparing fetch_insert_id: ".$dbh->errstr()."\n");
		}
		if (!$p->execute()) {
			die("Executing fetch_insert_id: ".$dbh->errstr()."\n");
		}
		if (!($r = $p->fetchrow_arrayref())) {
			die("Fetching the insert id: ".$dbh->errstr()."\n");
		}
		if (!defined($r->[0])) {
			die("An undefined insert ID fetched\n");
		}
		$p->finish();
		return $r->[0];
	} else {
		die("db_fetch_insert_id: no can do driver $cfg{db_driver}\n");
		return undef;
	}
}

sub qm_die($ $)
{
	my ($code, $msg) = @_;
	
	qm_syslog(LOG_ERR, $msg);
	exit($code);
}

sub qm_warn($)
{
	qm_syslog(LOG_WARNING, $_[0]);
}

sub qm_notify($)
{
	qm_syslog(LOG_INFO, $_[0]);
}

sub qm_debug($)
{
	qm_syslog(LOG_DEBUG, $_[0]);
}

sub qm_syslog($ $)
{
	my ($pri, $msg) = @_;

	eval {
		$msg =~ s/[\r\n]*$//;
		syslog($pri, "\%s\n", $msg);
	};
}
